const co = require('co');
const fs = require('co-fs');
const mkdirp = require('mkdirp');
const parseUrl = require('parseurl');
const path = require('path');
const resolve = path.resolve;
const send = require('send');
const sharp = require('sharp');

module.exports = function resizerMiddleware(options) {
  if (!options) {
    throw new Error('Options are required to run middleware');
  }

  if (!options.root) {
    throw new Error('"root" option is required');
  }

  if (!options.cache) {
    throw new Error('"cache" option is required');
  }

  const prefix = options.prefix || '';
  const transformNames = Object.keys(options.transforms || {});
  const allowedExtensions = options.allowedExtensions || ['jpg', 'jpeg', 'gif', 'png'];

  const regexp = new RegExp([
    '^(.*)',
    `/${prefix}(${transformNames.join('|')})`,
    '(/([^/]+)' + `\.(?:${allowedExtensions.join('|')}))(?:\.webp)?$`
  ].join(''), 'i');


  return function(req, res, next) {
    // Only GET + HEAD, if not NEXT()
    if (req.method != 'GET' && req.method != 'HEAD') {
      return next();
    }

    const pathname = parseUrl(req).pathname;
    const query = parseUrl(req).query;

    const match = pathname.match(regexp);

    if (!match) {
      return next();
    }

    const request = match.map(trimSlash);

    const fullRequest = request[0];
    const filePath = request[1]
    const transformName = request[2];
    const fileRequest = request[3];
    const fileName = request[4];

    const source = path.join(filePath, fileRequest);
    const destination = fullRequest;

    const resume = function resume(param, opts) {
      if (typeof param === 'string') {
        return send(req, param, {
          root: opts && opts.root || options.cache,
          maxAge: opts && opts.maxAge || options.maxAge || 31536000000,
          index: []
        }).pipe(res);
      }

      next(param);
    }

    const dst = resolve(options.cache, destination);
    const src = resolve(options.root, source);

    const getCachedFile = co.wrap(function *() {
      var cachedFile;

      const fileStats = () => fs.stat(dst);
      try {
        if (query && query.indexOf && query.indexOf('nocache') > -1) {
          yield fs.unlink(dst);
        }

        cachedFile = yield fileStats();

        res.setHeader('X-Hit-Cache', 1);
      } catch(err) {
        yield runTransform(src, dst, options.transforms[transformName]);

        cachedFile = yield fileStats();

        res.setHeader('X-Hit-Cache', 0);
      }

      if (cachedFile && cachedFile.isFile()) {
        return destination;
      }

      throw new Error('Could not process file.');
    })

    return co(function* () {
      try {
        return yield getCachedFile();
      } catch(err) {
        console.error(err)

        return [source, { root: options.root }];
      }
    }).then((data) => {
      if (!Array.isArray(data)) {
        data = [data];
      }

      resume.apply(null, data);
    }).catch(err => {
      console.error(err);
      resume(err);
    });
  }
}

function trimSlash(str) {
  return str.replace(/^\/|\/$/g, '');
}

function runTransform(from, to, handler) {
  const stream = handler(sharp(from), sharp);

  mkdirp.sync(path.dirname(to));

  return stream
    .toFile(to);
}
